/*
BSD 2-Clause License

Copyright (c) 2019, Beigesoft™
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package org.beigesoft.fct;

import java.util.Set;
import java.util.Map;
import java.util.HashMap;

import org.beigesoft.exc.ExcCode;
import org.beigesoft.rdb.IRdb;
import org.beigesoft.prc.IPrc;
import org.beigesoft.prc.About;
import org.beigesoft.prc.RefrI18n;
import org.beigesoft.prc.PrCsSmDtRo;

/**
 * <p>Factory of processors for base, secure non-transactional requests.</p>
 *
 * @param <RS> platform dependent record set type
 * @author Yury Demidenko
 */
public class FctPrcNtr<RS> implements IFctPrc {

  /**
   * <p>Main factory.</p>
   **/
  private FctBlc<RS> fctBlc;

  /**
   * <p>Outside factories.</p>
   **/
  private Set<IFctPrc> fctsPrc;

  //requested data:
  /**
   * <p>Processors map.</p>
   **/
  private final Map<String, IPrc> procs = new HashMap<String, IPrc>();

  /**
   * <p>Get processor in lazy mode (if bean is null then initialize it).</p>
   * @param pRvs request scoped vars
   * @param pPrNm - filler name
   * @return requested processor
   * @throws Exception - an exception
   */
  public final IPrc laz(final Map<String, Object> pRvs,
    final String pPrNm) throws Exception {
    IPrc rz = this.procs.get(pPrNm);
    if (rz == null) {
      synchronized (this) {
        rz = this.procs.get(pPrNm);
        if (rz == null) {
          if (About.class.getSimpleName().equals(pPrNm)) {
            rz = crPuAbout(pRvs);
          } else if (PrCsSmDtRo.class.getSimpleName().equals(pPrNm)) {
            rz = crPuPrCsSmDtRo(pRvs);
          } else if (RefrI18n.class.getSimpleName().equals(pPrNm)) {
            rz = crPuRefrI18n(pRvs);
          } else {
            if (this.fctsPrc != null) {
              for (IFctPrc fp : this.fctsPrc) {
                rz = fp.laz(pRvs, pPrNm);
                if (rz != null) {
                  break;
                }
              }
            }
            if (rz == null) {
              throw new ExcCode(ExcCode.WRCN, "There is no IProc: " + pPrNm);
            }
          }
        }
      }
    }
    return rz;
  }

  /**
   * <p>Create and put into the Map PrCsSmDtRo.</p>
   * @param pRvs request scoped vars
   * @return PrCsSmDtRo
   * @throws Exception - an exception
   */
  private PrCsSmDtRo crPuPrCsSmDtRo(
    final Map<String, Object> pRvs) throws Exception {
    PrCsSmDtRo rz = new PrCsSmDtRo();
    IFcCsvDrt fr = (IFcCsvDrt) this.fctBlc.laz(pRvs,
      IFcCsvDrt.class.getSimpleName());
    rz.setFctRet(fr);
    this.procs.put(PrCsSmDtRo.class.getSimpleName(), rz);
    this.fctBlc.lazLogStd(pRvs).info(pRvs, getClass(), PrCsSmDtRo.class
      .getSimpleName() + " has been created.");
    return rz;
  }

  /**
   * <p>Create and put into the Map RefrI18n.</p>
   * @param pRvs request scoped vars
   * @return RefrI18n
   * @throws Exception - an exception
   */
  private RefrI18n crPuRefrI18n(
    final Map<String, Object> pRvs) throws Exception {
    RefrI18n rz = new RefrI18n();
    rz.setHndI18nRq(this.fctBlc.lazHndI18nRq(pRvs));
    this.procs.put(RefrI18n.class.getSimpleName(), rz);
    this.fctBlc.lazLogStd(pRvs).info(pRvs, getClass(), RefrI18n.class
      .getSimpleName() + " has been created.");
    return rz;
  }

  /**
   * <p>Create and put into the Map About.</p>
   * @param pRvs request scoped vars
   * @return About
   * @throws Exception - an exception
   */
  private About<RS> crPuAbout(
    final Map<String, Object> pRvs) throws Exception {
    About<RS> rz = new About<RS>();
    @SuppressWarnings("unchecked")
    IRdb<RS> rdb = (IRdb<RS>) this.fctBlc.laz(pRvs, IRdb.class.getSimpleName());
    rz.setRdb(rdb);
    rz.setStgUvd(this.fctBlc.lazStgUvd(pRvs));
    this.procs.put(About.class.getSimpleName(), rz);
    this.fctBlc.lazLogStd(pRvs).info(pRvs, getClass(), About.class
      .getSimpleName() + " has been created.");
    return rz;
  }

  //Simple getters and setters:
  /**
   * <p>Getter for fctBlc.</p>
   * @return FctBlc<RS>
   **/
  public final synchronized FctBlc<RS> getFctBlc() {
    return this.fctBlc;
  }

  /**
   * <p>Setter for fctBlc.</p>
   * @param pFctBlc reference
   **/
  public final synchronized void setFctBlc(final FctBlc<RS> pFctBlc) {
    this.fctBlc = pFctBlc;
  }

  /**
   * <p>Getter for fctsPrc.</p>
   * @return Set<IFctPrc>
   **/
  public final synchronized Set<IFctPrc> getFctsPrc() {
    return this.fctsPrc;
  }

  /**
   * <p>Setter for fctsPrc.</p>
   * @param pFctsPrc reference
   **/
  public final synchronized void setFctsPrc(final Set<IFctPrc> pFctsPrc) {
    this.fctsPrc = pFctsPrc;
  }
}
